{{
  ****** Public Notes ******
  Modified by Duane Degn
  See name for last revision date.
  Started revision on December 20, 2012.

  My revisions may also be used in any way you see fit.

}}
{ ****** Private Notes ******

  121222a Start adding a code to use LEDs as a two dimentional
  array. The LEDs will be arranged with LED #0 in the top left corner
  and then moving down the array making it 8 LEDs high. LED #7 is
  located in the bottom left corner with LED #8 to the right of LED
  #7. LED #15 is to the right of LED #0.
  23b Works with 2D array.
  23c Add second strand.
  23e Works with two strands in 2D array.
  
}
{{
**  GE Color Effects Control Object
**  Version 0.2, 2010-12-02
**  by Robert Quattlebaum <darco@deepdarc.com>
**
**  http://www.deepdarc.com/2010/11/27/hacking-christmas-lights/
**  
**  This object file is public domain. You may use it as you see fit.
}}

CON { General constants }

  CMD_LEN = 26
  MAX_COLOR_VALUE = $F
  MAX_HUE = (MAX_COLOR_VALUE+1) * 6 - 1
  
  SECOND_STRAND_CHECK = LEDS_PER_STRAND << 20
   
  DEFAULT_INTENSITY = $CC     ' Default controler never uses larger than this.
  TOTAL_LEDS = LEDS_PER_STRAND * NUMBER_OF_STRANDS
  'BUFFER_SIZE = LEDS_PER_STRAND * NUMBER_OF_STRANDS                        
  'BROADCAST_BULB = $3F        ' Intensity set on this bulb affects all bulbs.
  'broadcast doesn't work now.

  STACK_SIZE = 100

  BUFFER_SIZE = ROWS * BUFFER_COLUMNS
    
CON { Color convenience constants }
  {COLOR_MASK          = $FFF
  COLOR_BLACK         = $000
  COLOR_WHITE         = $DDD  ' Default controler uses this value for white.
  COLOR_BLUE          = $F00
  COLOR_GREEN         = $0F0
  COLOR_RED           = $00F
  COLOR_CYAN          = COLOR_GREEN|COLOR_BLUE
  COLOR_MAGENTA       = COLOR_RED|COLOR_BLUE
  COLOR_YELLOW        = $07F 'COLOR_RED|COLOR_GREEN
  COLOR_ORANGE        = $03F
  COLOR_VIOLET        = $F08
  COLOR_INDIGO        = $F04
  COLOR_OTHER_WHITE   = $4DF        
  COLOR_YELLOW_GREEN  = $0BF
  COLOR_RED_VIOLET = $F0F
  COLOR_RED_ORANGE = $01F    }
CON

  RED_R = $00F
  RED_G = 0
  RED_B = 0
  ORANGE_R = $00F
  ORANGE_G = $030
  ORANGE_B = 0
  YELLOW_R = $00F
  YELLOW_G = $070
  YELLOW_B = 0
  GREEN_R = 0
  GREEN_G = $0F0
  GREEN_B = 0
  BLUE_R = 0
  BLUE_G = 0
  BLUE_B = $F00
  INDIGO_R = $004
  INDIGO_G = 0
  INDIGO_B = $F00
  VIOLET_R = $009
  VIOLET_G = 0
  VIOLET_B = $F00
  WHITE_R = $00F
  WHITE_G = $0D0
  WHITE_B = $400
  
  COLOR_MASK = INTENSITY | $FFF
  COLOR_BLACK = INTENSITY | $000
  COLOR_WHITE = INTENSITY | $DDD  ' Default controler uses this value for white.
  COLOR_BLUE = INTENSITY | $F00
  COLOR_GREEN = INTENSITY | $0F0
  COLOR_RED = INTENSITY | $00F
  COLOR_CYAN = INTENSITY | COLOR_GREEN|COLOR_BLUE
  COLOR_MAGENTA = COLOR_RED|COLOR_BLUE
  COLOR_YELLOW = INTENSITY | $07F 'COLOR_RED|COLOR_GREEN
  COLOR_ORANGE = INTENSITY | $03F
  COLOR_VIOLET = INTENSITY | $90F
  COLOR_INDIGO = INTENSITY | $40F
  COLOR_OTHER_WHITE = INTENSITY | $4DF        
  COLOR_YELLOW_GREEN  = INTENSITY | $0BF
  COLOR_RED_VIOLET = INTENSITY | $F0F
  COLOR_RED_ORANGE = INTENSITY | $01F
  COLOR_DIM_GREEN = ((DEFAULT_INTENSITY >> 3) << 12) | $0F0
  INTENSITY = DEFAULT_INTENSITY << 12
CON

  #0, NO_ACTION, UPDATE_SINGLE, UPDATE_ALL, UPDATE_2D
  
CON { These constants are needed only for the stand alone test mode. }
  _CLKMODE = XTAL1 + PLL16X 
  _XINFREQ = 5_000_000


CON

  MAX_INTENSITY = DEFAULT_INTENSITY '$FF
   
  MAX_BULB = 99 '49 '$3E              ' Theoretical last possible bulb address. Note
                              ' that normally there are only 50 bulbs on a string.
  LEDS_IN_STRAND = MAX_BULB + 1
  LEDS_PER_STRAND = 50
  NUMBER_OF_STRANDS = 2
  MAX_SHIFT_AMOUNT = LEDS_PER_STRAND / 2
  
  ROWS = 8
  DISPLAY_COLUMNS = 12
  BUFFER_COLUMNS = 13           ' one more than display columns to hold partial column

VAR

  long cog, command, singleCommand, bufferPtrHub
  long bufferOfPtrs[TOTAL_LEDS]
  long buffer[BUFFER_SIZE]
  long stack[STACK_SIZE], tempBuffer[MAX_SHIFT_AMOUNT]
  long millisecond
   
PUB stand_alone_test | localIndex, localIndex2
{{ Stand alone test. Used only when you run this object directly. }}
    ' Start up our cog.
    waitcnt(clkfreq * 4 + cnt)

    startTwo(12, 13, @buffer)
    waitcnt(clkfreq / 4 + cnt)

    ' Give the attached string the default bulb enumeration.
    set_standard_enum
    waitcnt(clkfreq / 4 + cnt)
    LoadBufferOfPtrs
   
    repeat
      VerticalStripes
      waitcnt(clkfreq / 2 + cnt)
      repeat 100
        Rotate2DHorizontal(1)
        waitcnt(clkfreq / 10 + cnt)
      HorizontalStripes 
      waitcnt(clkfreq / 2 + cnt)
      repeat 100
        Rotate2DVert(1)
        waitcnt(clkfreq / 10 + cnt)
    
    RefreshBuffer2D

    repeat
    
    repeat
      repeat result from 512 to 32
        Rotate(1, 1, result)
      repeat result from 32 to 512
        Rotate(1, 1, result)
      
    Rotate(1, 50, 256)
    Rotate(1, 50, 128)
    Rotate(1, 100, 64)
    'Rotate(1, 100, 32)
    
    repeat


PUB Start(pin, localPtr)
{{ Starts the cog for the object using the given pin for output. }}
   
  StartTwo(pin, -1, localPtr)
     
  return cog

PUB StartTwo(pin0, pin1, localPtr)
{{ Starts the cog for the object using the given pins for output. }}
    ' Stop the object if it is already running
  stop
  bufferPtrHub := localPtr
  ' Set up pin
  outputMask0 := |< pin0
  if pin1 <> -1
    outputMask1 := |< pin1
   
  ' Set the timing information
  baseComPeriod := (clkfreq/1_000_000) * 4
  millisecond := clkfreq/1_000
  ' Start the cog
  cog := cognew(@cog_init, @command) + 1
  waitcnt(clkfreq / 100 + cnt)
  if cog
    set_standard_enum
     
  return cog

PUB Stop
{{ Stops the object. You must call start again before you can use this object. }}
    if cog
        cogstop(cog~ - 1)       ' cog variable is cleared by this statement
    command~

PUB Rotate(direction, timesToShift, localDelay)

  localDelay #>= 1
  localDelay *= millisecond
  direction <#= MAX_SHIFT_AMOUNT
  direction #>= -MAX_SHIFT_AMOUNT
    repeat timesToShift
      ShiftBuffer(direction)
      waitcnt(localDelay + cnt)
      
PUB ShiftBuffer(direction)

  if direction > 0
    direction <#= MAX_SHIFT_AMOUNT
    longmove(@tempBuffer, bufferPtrHub, direction)
    'longmove(@tempBuffer, @buffer, direction)
    longmove(bufferPtrHub, bufferPtrHub + (4 * direction), BUFFER_SIZE - direction)
    'longmove(@buffer, @buffer + (4 * direction), BUFFER_SIZE - direction)
    longmove(bufferPtrHub + (4 * (BUFFER_SIZE - direction)), @tempBuffer, direction)
    'longmove(@buffer + (4 * (BUFFER_SIZE - direction)), @tempBuffer, direction)
   
  else
    direction := ||direction
    direction <#= MAX_SHIFT_AMOUNT
    longmove(@tempBuffer, bufferPtrHub + (4 * (BUFFER_SIZE - direction)), direction)
    'longmove(@tempBuffer, @buffer + (4 * (BUFFER_SIZE - direction)), direction)
    longmove(bufferPtrHub + (4 * direction), bufferPtrHub, BUFFER_SIZE - direction)
    'longmove(@buffer + (4 * direction), @buffer, BUFFER_SIZE - direction)
    longmove(bufferPtrHub, @tempBuffer, direction)
    'longmove(@buffer, @tempBuffer, direction)
  RefreshBuffer
     
PUB RefreshBuffer

  command := UPDATE_ALL
  repeat while command
  
PUB RefreshBuffer2D

  command := UPDATE_2D
  repeat while command
  
PUB set_standard_enum | i
{{ Performs the standard bulb enumeration for individual bulb control. }} 
  repeat i from 0 to MAX_BULB
    setBulb(i,DEFAULT_INTENSITY,COLOR_BLACK)
    waitcnt(clkfreq / 100 + cnt)
        
PUB sendRawCommand(x)
{{ Sends a single command. }} 
   

    ' Set the command. The cog will notice this
    ' and send it automatically, clearing
    ' the command variable when it is finished.
    singleCommand := x
    command := UPDATE_SINGLE

PUB GetRedPortion(color)

  result := color & $F
  
PUB GetGreenPortion(color)

 result := (color >> 4) & $F
 
PUB GetBluePortion(color)

  result := (color >> 8) & $F
  
PUB GetIntensityPortion(color)

  result := (color >> 12) & $FF

PUB MakeColor(redPortion, greenPortion, bluePortion, intensityPortion)

  result := ((intensityPortion & $FF) << 12) | ((bluePortion & $F) << 8) | ((greenPortion & $F) << 4) | (redPortion & $F)
       
PUB make_color_rgb(r,g,b)
{{ Helper function for converting discrete red, green, and blue values into a single 12 bit color value. }} 
    return (r&MAX_COLOR_VALUE)|((g&MAX_COLOR_VALUE)<<4)|((b&MAX_COLOR_VALUE)<<8)

PUB make_color_hue(hue)
{{ Helper function for creating a color based on a hue value. hue must be between 0 and MAX_HUE. }} 
    case (hue>>4)
        0:  hue &= MAX_COLOR_VALUE
            return make_color_rgb(MAX_COLOR_VALUE,hue,0)
        1:  hue &= MAX_COLOR_VALUE
            return make_color_rgb((MAX_COLOR_VALUE-hue),MAX_COLOR_VALUE,0)
        2:  hue &= MAX_COLOR_VALUE
            return make_color_rgb(0,MAX_COLOR_VALUE,hue)
        3:  hue &= MAX_COLOR_VALUE
            return make_color_rgb(0,(MAX_COLOR_VALUE-hue),MAX_COLOR_VALUE)
        4:  hue &= MAX_COLOR_VALUE
            return make_color_rgb(hue,0,MAX_COLOR_VALUE)
        5:  hue &= MAX_COLOR_VALUE
            return make_color_rgb(MAX_COLOR_VALUE,0,(MAX_COLOR_VALUE-hue))
    return 0

PUB setBulb(bulb, localIntensity, color) | x
{{ Convenience function for setting the intensity and color for the specified bulb index. }}
    'sendRawCommand(((bulb&$3F)<<20)|((intensity&$FF)<<12)|(color&COLOR_MASK))
  sendRawCommand(((bulb) << 20)|((localIntensity & $FF) << 12)|(color & COLOR_MASK))
  ' The buld ID may be above $3F if more than one strand is used.
  
PUB LoadBufferOfPtrs | localId
{{ Useful if you want the buffer containing the color values to be ordered
   differently than the way the bulbs are physically arranged.
   The array "bufferIds" will need to be filled with the bulb IDs
   as they are phsically arranged. }}
   
  repeat result from 0 to BUFFER_SIZE - 1 ' TOTAL_LEDS - 1
    if bufferIds[result] < TOTAL_LEDS
      bufferOfPtrs[bufferIds[result]] := bufferPtrHub + (4 * result)

     
PUB HorizontalStripes
{{ Assumes a display buffer 13 by 8.}}

  result := COLOR_RED
  'longfill(@buffer, result, 13)
  longfill(bufferPtrHub, result, 13)  
  result := COLOR_ORANGE
  longfill(bufferPtrHub + (13 * 4), result, 13)
  result := COLOR_YELLOW
  longfill(bufferPtrHub + (2 * 13 * 4), result, 13)
  result := COLOR_GREEN
  longfill(bufferPtrHub + (3 * 13 * 4), result, 13)
  result := COLOR_CYAN
  longfill(bufferPtrHub + (4 * 13 * 4), result, 13)
  result := COLOR_BLUE
  longfill(bufferPtrHub + (5 * 13 * 4), result, 13)
  result := COLOR_INDIGO
  longfill(bufferPtrHub + (6 * 13 * 4), result, 13)
  result := COLOR_VIOLET
  longfill(bufferPtrHub + (7 * 13 * 4), result, 13)
  result := COLOR_WHITE
   
  RefreshBuffer2D

PUB VerticalStripes | localIndex
{{ Assumes a display buffer 13 by 8.}}

  localIndex := 0
  result := COLOR_RED
  long[bufferPtrHub][localIndex++] := result
  result := COLOR_RED_ORANGE
  long[bufferPtrHub][localIndex++] := result
  result := COLOR_ORANGE
  long[bufferPtrHub][localIndex++] := result
  result := COLOR_YELLOW
  long[bufferPtrHub][localIndex++] := result 
  result := COLOR_YELLOW_GREEN
  long[bufferPtrHub][localIndex++] := result 
  result := COLOR_GREEN
  long[bufferPtrHub][localIndex++] := result 
  result := COLOR_CYAN
  long[bufferPtrHub][localIndex++] := result 
  result := COLOR_BLUE
  long[bufferPtrHub][localIndex++] := result 
  result := COLOR_INDIGO
  long[bufferPtrHub][localIndex++] := result 
  result := COLOR_VIOLET
  long[bufferPtrHub][localIndex++] := result 
  result := COLOR_RED_VIOLET
  long[bufferPtrHub][localIndex++] := result 
  result := COLOR_WHITE
  long[bufferPtrHub][localIndex++] := result 
  result := COLOR_WHITE
  repeat localIndex from 1 to 7
    longmove(bufferPtrHub + (4 * BUFFER_COLUMNS * localIndex), bufferPtrHub, BUFFER_COLUMNS) 
  RefreshBuffer2D

PUB Rotate2DHorizontal(direction) | localIndex

  if direction > 0
    repeat localIndex from 0 to ROWS - 1
      tempBuffer[localIndex] := long[bufferPtrHub][localIndex * BUFFER_COLUMNS]
      longmove(bufferPtrHub + (4 * localIndex * BUFFER_COLUMNS), bufferPtrHub + {
      } (4 * localIndex * BUFFER_COLUMNS) + 4, BUFFER_COLUMNS - 1)
      long[bufferPtrHub][(localIndex * BUFFER_COLUMNS) + BUFFER_COLUMNS - 1] := {
      } tempBuffer[localIndex] 
  else
    repeat localIndex from 0 to ROWS - 1
      tempBuffer[localIndex] := long[bufferPtrHub][(localIndex * BUFFER_COLUMNS) + BUFFER_COLUMNS - 1]
      longmove(bufferPtrHub + (4 * localIndex * BUFFER_COLUMNS) + 4, bufferPtrHub + (4 * localIndex * BUFFER_COLUMNS), BUFFER_COLUMNS - 1)
      long[bufferPtrHub][(localIndex * BUFFER_COLUMNS)] := tempBuffer[localIndex] 
  RefreshBuffer2D
  
PUB Rotate2DVert(direction)

  if direction > 0
    longmove(@tempBuffer, bufferPtrHub, BUFFER_COLUMNS)
    longmove(bufferPtrHub, bufferPtrHub + (4 * BUFFER_COLUMNS), BUFFER_SIZE - BUFFER_COLUMNS)
    longmove(bufferPtrHub + (4 * (BUFFER_SIZE - BUFFER_COLUMNS)), @tempBuffer, BUFFER_COLUMNS)
  else
    longmove(@tempBuffer, bufferPtrHub + (4 * (BUFFER_SIZE - BUFFER_COLUMNS)), BUFFER_COLUMNS)
    longmove(bufferPtrHub + (4 * BUFFER_COLUMNS), bufferPtrHub, BUFFER_SIZE - BUFFER_COLUMNS)
    longmove(bufferPtrHub, @tempBuffer, BUFFER_COLUMNS)
  RefreshBuffer2D
  
DAT
bufferIds               byte  0, 15, 16, 31, 32, 47, 48, 63, 64, 79, 80, 95, 96 
                        byte  1, 14, 17, 30, 33, 46, 49, 62, 65, 78, 81, 94, 97
                        byte  2, 13, 18, 29, 34, 45, 50, 61, 66, 77, 82, 93, 98
                        byte  3, 12, 19, 28, 35, 44, 51, 60, 67, 76, 83, 92, 99
                        byte  4, 11, 20, 27, 36, 43, 52, 59, 68, 75, 84, 91, 255 
                        byte  5, 10, 21, 26, 37, 42, 53, 58, 69, 74, 85, 90, 255 
                        byte  6,  9, 22, 25, 38, 41, 54, 57, 70, 73, 86, 89, 255 
                        byte  7,  8, 23, 24, 39, 40, 55, 56, 71, 72, 87, 88, 255 

                        
DAT { Cog Implementation }
                        org

cog_init                or      dira, outputMask0   ' Initialize output pin direction
                        or      dira, outputMask1
                        mov     singlePtr, par
                        add     singlePtr, #4
                        mov     arrayPtr, singlePtr
                        add     arrayPtr, #4
                        mov     ptrBuffer, arrayPtr
                        rdlong  arrayPtr, arrayPtr
                        add     ptrBuffer, #4
                        
loop                    wrlong  zero,par                ' Clear out the previous command

waitforcmd              rdlong  commandFromHub,par wz  ' Wait for next command
              if_z      jmp     #waitforcmd

                        ' Send the received command
                        add     commandFromHub, #jumpTable
                        jmp     commandFromHub
jumpTable               jmp     #loop
                        jmp     #sendCommand
                        jmp     #commandAll
                        jmp     #twoDArray
                        
sendCommand             rdlong  activeData, singlePtr
                        call    #sendDataRoutine
                         
                        ' Jump back to the start and wait for next command
                        jmp     #loop

commandAll              mov     countdown, ledsToCount
                        mov     tempPtr, arrayPtr
                        mov     bulbId, zero                                                                                                                                                                                                                                                                                                                                                                                                
arrayLoop               rdlong  value, tempPtr
                        and     value, twentyBits
                        mov     activeData, bulbId
                        shl     activeData, #20
                        add     activeData, value
                        call    #sendDataRoutine
                        add     bulbId, #1
                        add     tempPtr, #4
                        mov     waitTime, interElementPause
                        add     waitTime, cnt
                        waitcnt waitTime, interElementPause
                        djnz    countdown, #arrayLoop                                                                                                                                                                                                     
                        ' Jump back to the start and wait for next command
                        jmp     #loop

twoDArray               mov     countdown, ledsToCount
                        mov     tempPtr, ptrBuffer
                        mov     bulbId, zero                                                                                                                                                                                                                                                                                                                                                                                                
arrayLoop2D             rdlong  value, tempPtr
                        mov     activeData, bulbId
                        shl     activeData, #20
                        rdlong  value, value
                        and     value, twentyBits
                        add     activeData, value
                        call    #sendDataRoutine
                        add     bulbId, #1
                        add     tempPtr, #4
                        mov     waitTime, interElementPause
                        add     waitTime, cnt
                        waitcnt waitTime, interElementPause
                        djnz    countdown, #arrayLoop2D                                                                                                                                                                                                     
                        ' Jump back to the start and wait for next command
                        jmp     #loop

DAT { Send Command Subroutine }
' This routine expects the variable "activeData" to be loaded with the data
' to be sent on the data wire.
sendDataRoutine    ' Load number of data bits
                        cmp     activeData, strandCheck wc
                        ' if source is larger than d then c set
              if_c      mov     activeMask, outputMask0
              if_nc     mov     activeMask, outputMask1
              if_nc     sub     activeData, strandCheck               
                        mov     bitCountdown, #CMD_LEN
                         
                        ' Rotate the bits so that the next bit is at bit zero.
                        rol     activeData, #(32-CMD_LEN)
                         
                        ' Set up waitTime
                        mov     waitTime, baseComPeriod
                        add     waitTime, cnt
                         
                        ' Send start pulse
                        or      outa, activeMask   ' Output high
                        waitcnt waitTime, baseComPeriod  ' Wait 10 uSeconds
                        andn    outa, activeMask   ' Output low

outputLoop
                        rol     activeData, #1 wc

              if_c      waitcnt waitTime, baseComPeriod  ' Wait extra 10 uSeconds if one bit
                        waitcnt waitTime, baseComPeriod  ' Wait 10 uSeconds
                        or      outa, activeMask   ' Output
              if_nc     waitcnt waitTime, baseComPeriod  ' Wait extra 10 uSeconds if zero bit
                        waitcnt waitTime, baseComPeriod  ' Wait 10 uSeconds
                        andn    outa, activeMask   ' Output low
                         
                        ' Decrement bitCountdown ; jump if not Zero
                        djnz    bitCountdown, #outputLoop
                         
                        ' Finish up the frame with 30 uSecond quiet period.
                        waitcnt waitTime, baseComPeriod  ' Wait 10 uSeconds
                        waitcnt waitTime, baseComPeriod  ' Wait 10 uSeconds
                        waitcnt waitTime, baseComPeriod  ' Wait 10 uSeconds

sendDataRoutine_ret     ret

DAT { Constants }
zero                    long 0
twentyBits              long $FFFFF
interElementPause       long 1_600 '1_500 '1_400 ' (too short)  
baseComPeriod           long 0                  ' Set at init time
outputMask0             long 0                  ' Set at init time
outputMask1             long 0                  ' Set at init time
ledsToCount             long TOTAL_LEDS
strandCheck             long SECOND_STRAND_CHECK
DAT { Variables }
waitTime                res 1
commandFromHub          res 1
bitCountdown            res 1
singlePtr               res 1
arrayPtr                res 1
tempPtr                 res 1
ptrBuffer               res 1 
countdown               res 1                                                                                          
value                   res 1
bulbId                  res 1
activeData              res 1
activeMask              res 1
                                                                                                     